import boto3

import cv2

from aperturedb import CSVParser
from aperturedb.EntityUpdateDataCSV import SingleEntityUpdateDataCSV
from aperturedb.BlobNewestDataCSV import BlobNewestDataCSV
from aperturedb.SparseAddingDataCSV import SparseAddingDataCSV
from aperturedb.Sources import Sources
import logging
import os

logger = logging.getLogger(__name__)

HEADER_PATH = "filename"
HEADER_URL = "url"
HEADER_S3_URL = "s3_url"
HEADER_GS_URL = "gs_url"
PROPERTIES = "properties"
CONSTRAINTS = "constraints"
IMG_FORMAT = "format"


class ImageDataProcessor():
    """
    **Processing for Image data, used when loading images**
    """

    def __init__(self, check_image, n_download_retries):
        self.loaders = [self.load_image, self.load_url,
                        self.load_s3_url, self.load_gs_url]
        self.source_types = [HEADER_PATH,
                             HEADER_URL, HEADER_S3_URL, HEADER_GS_URL]

        self.check_image = check_image
        self.n_download_retries = n_download_retries

    def set_processor(self, use_dask: bool, source_type):
        self.source_type = source_type
        if use_dask == False and self.source_type == HEADER_S3_URL:
            # The connections by boto3 cause ResourceWarning. Known
            # issue: https://github.com/boto/boto3/issues/454
            s3_client = boto3.client('s3')
            self.sources = Sources(
                self.n_download_retries,
                s3_client=s3_client)
        else:
            self.sources = Sources(self.n_download_retries)

    def get_indices(self):
        return {
            "entity": {
                "_Image": self.get_indexed_properties()
            }
        }

    def load_image(self, filename):

        if self.check_image:
            try:
                a = cv2.imread(filename)
                if a.size <= 0:
                    logger.error(f"IMAGE SIZE ERROR: {filename}")
                    return False, None
            except Exception as e:
                logger.error(f"IMAGE ERROR: {filename}")
                logger.exception(e)

        return self.sources.load_from_file(filename)

    def check_image_buffer(self, img):

        if not self.check_image:
            return True

        try:
            # This is a full decoding of the image, a compute intensive operation.
            # decoded_img = cv2.imdecode(img, cv2.IMREAD_COLOR)
            # if decoded_img is None:
            #     logger.error("Error decoding image")
            #     print(img[0], img[1], img[2], img[3], img[4], img[5], img[6], img[7])
            #     return False

            # Instead, we do a much lighter check on the header of the image,
            # which is already in memory.
            # Check header signature for file format:
            # https://en.wikipedia.org/wiki/List_of_file_signatures

            # NOTE: Make sure we are checking for the same values
            #       as Image.cc on ApertureDB
            # Ideally, we don't want this logic here because it is replicated
            # on the server side, and we have to maintain it in both places.
            # But because it is expensive to send the image to the server
            # only to find out it is bad, we do it here.

            # JPEG
            if img[0] == 255 and img[1] == 216:
                return True
            # PNG
            if img[0] == 137 and img[1] == 80 and img[2] == 78 and img[3] == 71:
                return True
            # TIF
            if img[0] == 73 and img[1] == 73 and img[2] == 42:
                return True

        except Exception as e:
            logger.exception(e)
            return False

        return False  # When header is not recognized

    def load_url(self, url):
        return self.sources.load_from_http_url(url, self.check_image_buffer)

    def load_s3_url(self, s3_url):
        return self.sources.load_from_s3_url(s3_url, self.check_image_buffer)

    def load_gs_url(self, gs_url):
        return self.sources.load_from_gs_url(gs_url, self.check_image_buffer)


class ImageDataCSV(CSVParser.CSVParser, ImageDataProcessor):
    """**ApertureDB Image Data.**

    This class loads the Image Data which is present in a CSV file,
    and converts it into a series of ApertureDB queries.


    :::note Is backed by a CSV file with the following columns:

    ``filename``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

    OR

    ``url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

    OR

    ``s3_url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

    OR

    ``gs_url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``
    :::

    Example CSV file::

        filename,id,label,constraint_id,format
        /home/user/file1.jpg,321423532,dog,321423532,jpg
        /home/user/file2.jpg,42342522,cat,42342522,png
        ...

    Example usage:

    ``` python

        data = ImageDataCSV("/path/to/ImageData.csv")
        loader = ParallelLoader(client)
        loader.ingest(data)
    ```

    or

    ``` bash

        adb ingest from-csv /path/to/ImageData.csv --ingest-type IMAGE
    ```


    :::info
    In the above example, the constraint_id ensures that an Image with the specified
    id would be only inserted if it does not already exist in the database.
    :::
    """

    def __init__(self, filename: str, check_image: bool = True, n_download_retries: int = 3, **kwargs):

        ImageDataProcessor.__init__(
            self, check_image, n_download_retries)
        CSVParser.CSVParser.__init__(self, filename, **kwargs)

        source_type = self.header[0]
        self.set_processor(self.use_dask, source_type)

        self.format_given = IMG_FORMAT in self.header
        self.props_keys = [x for x in self.header[1:]
                           if not x.startswith(CSVParser.CONSTRAINTS_PREFIX)]
        self.props_keys = [
            x for x in self.props_keys if x != IMG_FORMAT]
        self.constraints_keys = [x for x in self.header[1:]
                                 if x.startswith(CSVParser.CONSTRAINTS_PREFIX)]

        if self.source_type not in self.source_types:
            logger.error("Source not recognized: " + self.source_type)
            raise Exception("Error loading image: " + filename)
        self.source_loader = {
            st: sl for st, sl in zip(self.source_types, self.loaders)
        }

        self.relative_path_prefix = os.path.dirname(self.filename) \
            if self.source_type == HEADER_PATH and self.blobs_relative_to_csv else ""

        self.command = "AddImage"

    def getitem(self, idx):
        idx = self.df.index.start + idx

        image_path = os.path.join(
            self.relative_path_prefix, self.df.loc[idx, self.source_type])
        img_ok, img = self.source_loader[self.source_type](image_path)

        if not img_ok:
            logger.error("Error loading image: " + image_path)
            raise Exception("Error loading image: " + image_path)

        q = []
        blobs = []
        custom_fields = {}
        if self.format_given:
            custom_fields["format"] = self.df.loc[idx, IMG_FORMAT]
        ai = self._basic_command(idx, custom_fields)
        # Each getitem query should be properly defined with a ref.
        # A ref shouldb be added to each of the commands from getitem implementation.
        # This is because a transformer or ref updater in the PQ
        # will need to know which command to update.
        ai[self.command]["_ref"] = 1
        blobs.append(img)
        q.append(ai)

        return q, blobs

    def get_indices(self):
        return {
            "entity": {
                "_Image": self.get_indexed_properties()
            }
        }

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] not in self.source_types:
            field = self.header[0]
            allowed = ", ".join(self.source_types)
            raise Exception(
                f"Error with CSV file field: {field}. Allowed values: {allowed}")


class ImageUpdateDataCSV(SingleEntityUpdateDataCSV, ImageDataProcessor):
    """
    **ApertureDB Image CSV Parser for Adding an Image and updating the properties on the image.**
    Usage is in EntityUpdateDataCSV.

    Note that this class will not change a blob in an existing entity. If looking to change a blob in
     an existing entity, look at ImageForceNewestDataCSV, but be careful of all caveats.

    """

    def __init__(self, filename: str, check_image: bool = True, n_download_retries: int = 3, **kwargs):
        ImageDataProcessor.__init__(
            self, check_image, n_download_retries)
        SingleEntityUpdateDataCSV.__init__(
            self, "Image", filename, **kwargs)

        source_type = self.header[0]
        self.set_processor(self.use_dask, source_type)

        # this class loads a blob, so must set the first query to have a blob.
        self.blobs_per_query = [1, 0]
        self.format_given = IMG_FORMAT in self.header
        self.props_keys = list(filter(lambda prop: prop not in [
                               IMG_FORMAT, "filename"], self.props_keys))
        if self.source_type not in self.source_types:
            logger.error("Source not recognized: " + self.source_type)
            raise Exception("Error loading image: " + filename)
        self.source_loader = {
            st: sl for st, sl in zip(self.source_types, self.loaders)
        }

    def getitem(self, idx):
        blob_set = []
        [query_set, empty_blobs] = super().getitem(idx)

        image_path = os.path.join(
            self.relative_path_prefix, self.df.loc[idx, self.source_type])
        img_ok, img = self.source_loader[self.source_type](image_path)
        if not img_ok:
            logger.error("Error loading image: " + image_path)
            raise Exception("Error loading image: " + image_path)
        # element has 2 queries, each with 1 blob.
        blob_set = [[img], []]
        # must wrap the blob return for this item in a list
        return [query_set, [blob_set]]

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] not in self.source_types:
            raise Exception(
                f"Error with CSV file field: {self.header[0]}. Must be first field")
        else:
            SingleEntityUpdateDataCSV.validate(self)


class ImageForceNewestDataCSV(BlobNewestDataCSV, ImageDataProcessor):
    """
    **ApertureDB Image CSV Parser for Maintaining a Blob set with changing blob data.**
    See BlobNewestDataCSV for usage.
    """

    def __init__(self, filename, check_image=True, n_download_retries=3, **kwargs):
        ImageDataProcessor.__init__(
            self, check_image, n_download_retries)
        BlobNewestDataCSV.__init__(
            self, "Image", filename, **kwargs)

        source_type = self.header[0]
        self.set_processor(self.use_dask, source_type)

        self.format_given = IMG_FORMAT in self.header
        self.props_keys = list(filter(lambda prop: prop not in [
                               IMG_FORMAT, "filename"], self.props_keys))
        if self.source_type not in self.source_types:
            logger.error("Source not recognized: " + self.source_type)
            raise Exception("Error loading image: " + filename)
        self.source_loader = {
            st: sl for st, sl in zip(self.source_types, self.loaders)
        }

    def read_blob(self, idx):
        relative_path_prefix = os.path.dirname(self.filename) \
            if self.source_type == HEADER_PATH else ""
        image_path = os.path.join(
            relative_path_prefix, self.df.loc[idx, self.source_type])
        img_ok, img = self.source_loader[self.source_type](image_path)
        if not img_ok:
            logger.error("Error loading image: " + image_path)
            raise Exception("Error loading image: " + image_path)
        return img


class ImageSparseAddDataCSV(SparseAddingDataCSV, ImageDataProcessor):
    """
    **ApertureDB Spare Loading Image Data.**

    ImageSparseAddDataCSV should be used the same as ImageDataCSV.

    See SparseAddingDataCSV for description of when to use this versus ImageDataCSV.


    Example CSV file::

        filename,id,label,constraint_id,format
        /home/user/file1.jpg,321423532,dog,321423532,jpg
        /home/user/file2.jpg,42342522,cat,42342522,png
        ...

    Example usage:

    ```python

        data = ImageSparseAddDataCSV("/path/to/ImageData.csv")
        loader = ParallelLoader(client)
        loader.ingest(data)
    ```
    """

    def __init__(self, filename: str, check_image: bool = True, n_download_retries: int = 3, **kwargs):
        ImageDataProcessor.__init__(
            self, check_image, n_download_retries)
        SparseAddingDataCSV.__init__(self, "Image", filename, **kwargs)
        source_type = self.header[0]
        self.set_processor(self.use_dask, source_type)

        self.format_given = IMG_FORMAT in self.header
        self.props_keys = list(filter(lambda prop: prop not in [
                               IMG_FORMAT, "filename"], self.props_keys))
        if self.source_type not in self.source_types:
            logger.error("Source not recognized: " + self.source_type)
            raise Exception("Error loading image: " + filename)
        self.source_loader = {
            st: sl for st, sl in zip(self.source_types, self.loaders)
        }

    def getitem(self, idx):
        blob_set = []
        [query_set, empty_blobs] = super().getitem(idx)

        image_path = os.path.join(
            self.relative_path_prefix, self.df.loc[idx, self.source_type])
        img_ok, img = self.source_loader[self.source_type](image_path)

        if not img_ok:
            logger.error("Error loading image: " + image_path)
            raise Exception("Error loading image: " + image_path)
        # element has 2 queries, only second has blob
        blob_set = [[], [img]]
        # must wrap the blob return for this item in a list
        return [query_set, [blob_set]]



import pytest
from aperturedb.Connector import Connector
from aperturedb.ConnectorRest import ConnectorRest
from aperturedb.ParallelLoader import ParallelLoader
import dbinfo
import pandas as pd

import logging
logger = logging.getLogger(__name__)


class TestBadResponses():

    def test_Error_code_2(self, db: Connector, insert_data_from_csv, monkeypatch):
        count = 0
        original_q = db._query

        def test_response_half_exist(a: Connector, query, blobs):
            nonlocal count
            if "AddImage" not in query[0]:
                count += 1
                resp = original_q(query, blobs)
                return resp
            response = []
            for i in range(len(query)):
                result = {"info": "Object Exists!",
                          "status": 2} if i % 2 == 0 else {"status": 0}
                response.append({"AddImage": result})

            return (response, [])
        monkeypatch.setattr(Connector, "_query", test_response_half_exist)
        monkeypatch.setattr(ParallelLoader, "get_existing_indices", lambda x: {
                            "entity": {"_Image": {"id"}}})
        data, loader = insert_data_from_csv(
            in_csv_file = "./input/images.adb.csv")
        assert loader.error_counter == 0
        assert loader.get_succeeded_queries() == len(data)
        assert loader.get_succeeded_commands() == len(data)

    def test_Error_code_3(self, db: Connector, insert_data_from_csv, monkeypatch):
        count = 0
        original_q = db._query

        def test_response_half_non_unique(a: Connector, query, blobs):
            nonlocal count
            if "AddImage" not in query[0]:
                count += 1
                resp = original_q(query, blobs)
                return resp
            response = None
            for i in range(len(query)):
                result = {
                    'info': 'JSON Command 1: expecting 1 but got 2', 'status': 3}
                response = result
                break

            return (response, [])
        monkeypatch.setattr(Connector, "_query", test_response_half_non_unique)
        monkeypatch.setattr(ConnectorRest, "_query",
                            test_response_half_non_unique)
        monkeypatch.setattr(ParallelLoader, "get_existing_indices", lambda x: {
                            "entity": {"_Image": {"id"}}})
        input_data = pd.read_csv("./input/images.adb.csv")
        data, loader = insert_data_from_csv(
            in_csv_file = "./input/images.adb.csv", expected_error_count = len(input_data))
        assert loader.error_counter == 0, f"Error counter: {loader.error_counter=}"
        assert loader.get_succeeded_queries(
        ) == 0, f"Queries: {loader.get_succeeded_queries()=}"
        assert loader.get_succeeded_commands(
        ) == 0, f"Commands: {loader.get_succeeded_commands()=}"

    def test_AuthFailure(self, monkeypatch):

        def failed_auth_query(conn_obj, ignored_query):
            # generate a response from the server which is not the expected Auth result.
            # _query returns the server response json and an array of blobs.
            return ({"info": "Internal Server Error 42", "status": -1, "ignored": ignored_query}, [])

        monkeypatch.setattr(Connector, "_query", failed_auth_query)

        with pytest.raises(Exception) as conn_exception:
            db = Connector(
                host = dbinfo.DB_TCP_HOST,
                port = dbinfo.DB_TCP_PORT,
                user = dbinfo.DB_USER,
                password = dbinfo.DB_PASSWORD,
                use_ssl = True)
            db.query([{
                "FindImage": {
                    "results": {
                        "limit": 5
                    }
                }
            }])

        assert "Unexpected response" in str(conn_exception.value)



"""
Common library functions for ApertureDB.
This will not have a big class structure, but rather a collection of functions
This is the place to put functions that are reused in codebase.
"""
import importlib
import math
import os
import sys
from typing import Any, Callable, Optional, Tuple, Dict, Union
import logging
import json

from aperturedb.Configuration import Configuration
from aperturedb.Connector import Connector
from aperturedb.ConnectorRest import ConnectorRest
from aperturedb.types import Blobs, CommandResponses, Commands

logger = logging.getLogger(__name__)


def import_module_by_path(filepath: str) -> Any:
    """
    This function imports a module given a path to a python file.
    """
    module_name = os.path.basename(filepath)[:-3]
    spec = importlib.util.spec_from_file_location(module_name, filepath)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


def __create_connector(configuration: Configuration):
    if configuration.use_rest:
        connector = ConnectorRest(
            host=configuration.host,
            port=configuration.port,
            user=configuration.username,
            password=configuration.password,
            use_ssl=configuration.use_ssl,
            config=configuration)
    else:
        connector = Connector(
            host=configuration.host,
            port=configuration.port,
            user=configuration.username,
            password=configuration.password,
            use_ssl=configuration.use_ssl,
            config=configuration)
    logger.debug(
        f"Created connector using: {configuration}. Will connect on query.")
    return connector


def _create_configuration_from_json(config: Union[Dict, str],
                                    name: Optional[str] = None, name_required: bool = False) -> Connector:
    """
    **Create a connector to the database from a JSON configuration.**

    Args:
        config (Dict or str): The configuration.

    Returns:
        Connector: The connector to the database.
    """
    if isinstance(config, str):
        try:
            config = json.loads(config)
        except json.JSONDecodeError as e:
            # Cannot print the JSON string because it may contain a password.
            assert False, f"problem decoding JSON config string: {e}"
    assert isinstance(
        config, dict), f"config must be a dict or a JSON object string: {type(config)}"

    # Remove the password from the configuration before logging.
    clean_config = {k: v for k, v in config.items() if k != "password"}

    # These fields are required.
    assert "host" in config, f"host is required in the configuration: {clean_config}"
    assert "username" in config, f"username is required in the configuration: {clean_config}"
    assert "password" in config, f"password is required in the configuration: {clean_config}"

    # These fields have no default in the Configuration class.
    if 'port' not in config:
        config["port"] = 55555

    if name is not None:
        config["name"] = name  # will overwrite the name in the config

    if name_required:
        assert "name" in config, f"name is required in the configuration: {clean_config}"
    elif 'name' not in config:
        config["name"] = "from_json"

    configuration = Configuration(**config)
    return configuration


def _get_colab_secret(name: str) -> Optional[str]:
    try:
        from google.colab import userdata
        return userdata.get(name)
    except ImportError:  # Not in Colab environment
        return None
    except AttributeError:  # In Colab environment but not in a notebook
        logger.debug(
            "In Colab environment but not in a notebook. Cannot read secrets.")
        return None
    except userdata.NotebookAccessError:  # Permission to access secrets not granted
        logger.debug(
            "Permission to access secrets not granted to this notebook.")
        return None
    except userdata.SecretNotFoundError:  # This secret does not exist
        logger.debug(
            f"Secret '{name}' not found in Google Colab.")
        return None
    except Exception as e:  # Unexpected error
        logger.error(
            f"Unexpected error while reading secret '{name}' from Google Colab: {e}")
        return None


def _get_dotenv_secret(name: str) -> Optional[str]:
    try:
        from dotenv import dotenv_values
        config = dotenv_values(".env")
        return config.get(name)
    except ImportError:  # dotenv not installed
        logger.warning(
            "dotenv not installed. Cannot read secrets from .env file.")
        return None


def _store_config(config: Configuration, name: str):
    logger.info(
        f"Storing and activating configuration '{name}' in the global configuration: {config}")

    from aperturedb.cli.configure import create
    create(
        name=name,
        host=config.host,
        port=config.port,
        username=config.username,
        password=config.password,
        use_ssl=config.use_ssl,
        use_rest=config.use_rest,
        interactive=False,
        overwrite=True,
        active=True,
    )


def create_connector(
    name: Optional[str] = None,
    create_config_for_colab_secret=True,
) -> Connector:
    """
    **Create a connector to the database.**

    This function chooses a configuration in the folowing order:
    1. The configuration named by the `name` parameter.
    2. The configuration described in the `APERTUREDB_JSON` environment variable.
    3. The configuration described in the `APERTUREDB_JSON` Google Colab secret.
    4. The configuration described in the `APERTUREDB_JSON` secret in a `.env` file.
    5. The configuration named by the `APERTUREDB_CONFIG` environment variable.
    6. The active configuration.

    If there are both global and local configurations with the same name, the global configuration is preferred.

    See :ref:`adb config <adb_config>`_ command-line tool for more information.

    Arguments:
        name (str, optional): The name of the configuration to use. Default is None.
        create_config_for_colab_secret (bool, optional): Whether to create a configuration from the Google Colab secret. Default is True.

    Returns:
        Connector: The connector to the database.

    Note about Google Colab secret: This secret is available in the context of a notebook running on Google Colab.  In particular, it is not available to the `adb` CLI tool running in a Colab notebook or any scripts run within a notebook.  To resolve this issue, a configuration is automatically created and activated in this case.  Use the `create_config_for_colab_secret` parameter to disable this behavior.
    """
    from aperturedb.cli.configure import ls
    _all_configs = None

    def get_all_configs():
        nonlocal _all_configs
        if _all_configs is None:
            _all_configs = ls(log_to_console=False)
        return _all_configs

    def lookup_config_by_name(name: str, source: str) -> Configuration:
        configs = get_all_configs()

        if "global" in configs and name in configs["global"]:
            return configs["global"][name]
        if "local" in configs and name in configs["local"]:
            return configs["local"][name]
        assert False, f"Configuration '{name}' not found ({source})."

    if name is not None:
        logger.info(f"Using configuration '{name}' explicitly")
        config = lookup_config_by_name(name, "explicit")
    elif (data := os.environ.get("APERTUREDB_JSON")) is not None and data != "":
        logger.info(
            f"Using configuration from APERTUREDB_JSON environment variable")
        config = _create_configuration_from_json(data)
    elif (data := _get_colab_secret("APERTUREDB_JSON")) is not None and data != "":
        logger.info(
            f"Using configuration from APERTUREDB_JSON Google Colab secret")
        config = _create_configuration_from_json(data)
        if create_config_for_colab_secret:
            logger.info(
                f"Creating and activating configuration from APERTUREDB_JSON Google Colab secret")
            _store_config(config, 'google_colab')
    elif (data := _get_dotenv_secret("APERTUREDB_JSON")) is not None and data != "":
        logger.info(
            f"Using configuration from APERTUREDB_JSON secret in .env file")
        config = _create_configuration_from_json(data)
    elif (name := os.environ.get("APERTUREDB_CONFIG")) is not None and name != "":
        logger.info(
            f"Using configuration '{name}' from APERTUREDB_CONFIG environment variable")
        config = lookup_config_by_name(name, "envar")
    elif "active" in get_all_configs():
        name = get_all_configs()["active"]
        config = lookup_config_by_name(name, "active")
        logger.info(f"Using active configuration '{name}'")
    else:
        assert False, "No configuration found."
    logger.info(f"Configuration: {config}")
    return __create_connector(config)


def execute_query(client: Connector, query: Commands,
                  blobs: Blobs = [],
                  success_statuses: list[int] = [0],
                  response_handler: Optional[Callable] = None, commands_per_query: int = 1, blobs_per_query: int = 0,
                  strict_response_validation: bool = False, cmd_index=None) -> Tuple[int, CommandResponses, Blobs]:
    """
    Execute a batch of queries, doing useful logging around it.
    Calls the response handler if provided.

    This should be used (without the parallel machinery) instead of
    Connector.query to keep the response handling consistent, better logging, etc.

    Args:
        client (Connector): The database connector.
        query (Commands): List of commands to execute.
        blobs (Blobs, optional): List of blobs to send.
        success_statuses (list[int], optional): The list of success statuses. Defaults to [0].
        response_handler (Callable, optional): The response handler. Defaults to None.
        commands_per_query (int, optional): The number of commands per query. Defaults to 1.
        blobs_per_query (int, optional): The number of blobs per query. Defaults to 0.
        strict_response_validation (bool, optional): Whether to strictly validate the response. Defaults to False.

    Returns:
        int: The result code.
            - 0 : if all commands succeeded
            - 1 : if there was -1 in the response
            - 2 : For any other code.
        CommandResponses: The response.
        Blobs: The blobs.
    """
    result = 0
    logger.debug(f"Query={query}")
    r, b = client.query(query, blobs)
    logger.debug(f"Response={r}")

    if client.last_query_ok():
        if response_handler is not None:
            try:
                map_response_to_handler(response_handler,
                                        query, blobs, r, b, commands_per_query, blobs_per_query,
                                        cmd_index)
            except BaseException as e:
                logger.exception(e)
                if strict_response_validation:
                    raise e
    else:
        # Transaction failed entirely.
        logger.error(f"Failed query = {query} with response = {r}")
        result = 1

    statuses = {}
    if isinstance(r, dict):
        statuses[r['status']] = [r]
    elif isinstance(r, list):
        # add each result to a list of the responses, keyed by the response
        # code.
        [statuses.setdefault(result[cmd]['status'], []).append(result)
         for result in r for cmd in result]
    else:
        logger.error("Response in unexpected format")
        result = 1

    # last_query_ok means result status >= 0
    if result != 1:
        warn_list = []
        for status, results in statuses.items():
            if status not in success_statuses:
                for wr in results:
                    warn_list.append(wr)
        if len(warn_list) != 0:
            logger.warning(
                f"Partial errors:\r\n{json.dumps(query)}\r\n{json.dumps(warn_list)}")
            result = 2

    return result, r, b


def map_response_to_handler(handler, query, query_blobs,  response, response_blobs,
                            commands_per_query, blobs_per_query, cmd_index_offset):
    # We could potentially always call this handler function
    # and let the user deal with the error cases.
    blobs_returned = 0
    for i in range(math.ceil(len(query) / commands_per_query)):
        start = i * commands_per_query
        end = start + commands_per_query
        blobs_start = i * blobs_per_query
        blobs_end = blobs_start + blobs_per_query

        b_count = 0
        if issubclass(type(response), list):
            for req, resp in zip(query[start:end], response[start:end]):
                for k in req:
                    blob_returning_commands = ["FindImage", "FindBlob", "FindVideo",
                                               "FindDescriptor", "FindBoundingBox"]
                    if k in blob_returning_commands and "blobs" in req[k] and req[k]["blobs"]:
                        count = resp[k]["returned"]
                        b_count += count

        # The returned blobs need to be sliced to match the
        # returned entities per command in query.
        handler(
            query[start:end],
            query_blobs[blobs_start:blobs_end],
            response[start:end] if issubclass(
                type(response), list) else response,
            response_blobs[blobs_returned:blobs_returned + b_count] if
            len(response_blobs) >= blobs_returned + b_count else None,
            None if cmd_index_offset is None else cmd_index_offset + i)
        blobs_returned += b_count


def issue_deprecation_warning(old_name, new_name):
    """
    Issue a deprecation warning for a function and class.
    """
    logger.warning(
        f"{old_name} is deprecated and will be removed in a future release. Use {new_name} instead.")




from aperturedb import ParallelQuery
from aperturedb.Connector import Connector
from aperturedb.Utils import Utils
from aperturedb.Subscriptable import Subscriptable

import numpy as np
import logging
logger = logging.getLogger(__name__)


class ParallelLoader(ParallelQuery.ParallelQuery):
    """
    **Parallel and Batch Loader for ApertureDB**

    This takes a dataset (which is a collection of homogeneous objects)
    or a derived class, and optimally inserts them into database by splitting them
    into batches, and passing the batches to multiple workers.
    """

    def __init__(self, client: Connector, dry_run: bool = False):
        super().__init__(client, dry_run=dry_run)
        self.utils = Utils(self.client)
        self.type = "element"

    def get_existing_indices(self):
        schema = self.utils.get_schema()
        existing_indices = {}
        if schema:
            for index_type in (("entity", "entities"), ("connection", "connections")):
                foo = schema.get(index_type[1]) or {}
                bar = foo.get("classes") or {}
                for cls_name, cls_schema in bar.items():
                    props = cls_schema.get("properties") or {}
                    for prop_name, prop_schema in props.items():
                        if prop_schema[1]:  # indicates property has an index
                            existing_indices.setdefault(index_type[0], {}).setdefault(
                                cls_name, set()).add(prop_name)
        return existing_indices

    def query_setup(self, generator: Subscriptable) -> None:
        """
        Runs the setup for the loader, which includes creating indices
        Currently, it only creates indices for the properties that are
        also used for constraint.

        Will only run when the argument generator has a get_indices method that returns
        a dictionary of the form:

        ``` python
        {
            "entity": {
                "class_name": ["property_name"]
            },
        }

        or

        {
            "connection": {
                "class_name": ["property_name"]
            },
        }
        ```

        Args:
            generator (Subscriptable): The Subscriptable object that is being ingested
        """
        if hasattr(generator, "get_indices"):
            schema = self.utils.get_schema()

            indices_needed = generator.get_indices()
            for schema_type, schema_type_plural in [("entity", "entities"), ("connection", "connections")]:
                for entity_class in indices_needed.get(schema_type, {}):
                    for property_name in indices_needed[schema_type][entity_class]:
                        schema_type = schema.get(schema_type_plural, {}) or {}
                        if property_name not in schema_type.get('classes', {}).get(entity_class, {}).get('properties', {}):
                            if not self.utils.create_entity_index(entity_class, property_name):
                                logger.warning(
                                    f"Failed to create index for {entity_class}.{property_name}")
                        else:
                            logger.info(
                                f"Index for {entity_class}.{property_name} already exists")

    def ingest(self, generator: Subscriptable, batchsize: int = 1, numthreads: int = 4, stats: bool = False) -> None:
        """
        **Method to ingest data into the database**

        Args:
            generator (Subscriptable): The list of data, or a class derived from [Subscriptable](/python_sdk/helpers/Subscriptable) to be ingested.
            batchsize (int, optional): The size of batch to be used. Defaults to 1.
            numthreads (int, optional): Number of workers to create. Defaults to 4.
            stats (bool, optional): If stats need to be presented, realtime. Defaults to False.
        """
        logger.info(
            f"Starting ingestion with batchsize={batchsize}, numthreads={numthreads}")
        self.query(generator, batchsize, numthreads, stats)

    def print_stats(self) -> None:

        times = np.array(self.get_times())
        total_queries_exec = len(times)

        print("============ ApertureDB Loader Stats ============")
        print(f"Total time (s): {self.total_actions_time}")
        print(f"Total queries executed: {total_queries_exec}")

        succeeded_queries = sum([stat["succeeded_queries"]
                                for stat in self.actual_stats])
        succeeded_commands = sum([stat["succeeded_commands"]
                                  for stat in self.actual_stats])

        if succeeded_queries == 0:
            print("All queries failed!")

        else:
            mean = np.mean(times)
            std = np.std(times)
            tp = 0 if mean == 0 else 1 / mean * self.numthreads

            print(f"Avg Query time (s): {mean}")
            print(f"Query time std: {std}")
            print(f"Avg Query Throughput (q/s): {tp}")

            i_tp = self.total_actions / self.total_actions_time
            print(
                f"Overall insertion throughput ({self.type}/s): {i_tp if self.error_counter == 0 else 'NaN'}")

            if self.error_counter > 0:
                err_perc = 100 * self.error_counter / total_queries_exec
                print(f"Total errors encountered: {self.error_counter}")
                print(f"Errors (%): {err_perc}")

            # TODO this does not take into account that the last
            # batch may be smaller than batchsize
            print(f"Total inserted elements: {succeeded_queries}")
            print(f"Total successful commands: {succeeded_commands}")
        print("=================================================")



"""
**Data Model Classes to support (pydantic) model based ingestiton.**
"""
from __future__ import annotations
from pydantic import BaseModel, Field
from typing_extensions import Annotated, List
from typing import ClassVar, Optional
from uuid import uuid4
from aperturedb.Query import ObjectType, PropertyType, RangeType


class IdentityDataModel(BaseModel):
    """Base class for all entities in ApertureDB.
    Generates a default UUID for the entity.
    """

    id: Annotated[str, Field(default_factory=lambda: uuid4().hex)]
    # Change as per the docs for the error
    # https://docs.pydantic.dev/dev-v2/usage/errors/#model-field-overridden
    type: ClassVar[ObjectType] = ObjectType.ENTITY


class BlobDataModel(IdentityDataModel):
    """Base class for all blob entities in ApertureDB.
    """
    url: Annotated[str, Field(
        title="URL", description="URL to file, http, s3 or gs resource")]
    type = ObjectType.BLOB


class ImageDataModel(BlobDataModel):
    """Base class for all image objects in ApertureDB.
    """
    type = ObjectType.IMAGE


class ClipDataModel(IdentityDataModel):
    """Base class for all clip objects in ApertureDB.
    """
    type = ObjectType.CLIP
    range_type: Annotated[RangeType,
                          Field(title="Range Type", description="Range type",
                                default=RangeType.TIME),
                          PropertyType.SYSTEM]
    start: Annotated[float, Field(title="Start", description="Start point as frame, time(hh:mm:ss.uuuuuu) or fraction"),
                     PropertyType.SYSTEM]
    stop: Annotated[float, Field(title="Stop", description="Stop point as frame, time(hh:mm:ss.uuuuuu) or fraction"),
                    PropertyType.SYSTEM]


class VideoDataModel(BlobDataModel):
    """Data model for video objects in ApertureDB.
    """
    type = ObjectType.VIDEO


class DescriptorDataModel(IdentityDataModel):
    """Descriptor (Embedding) data model for ApertureDB.
    """
    type = ObjectType.DESCRIPTOR
    vector: Annotated[List[float], Field(
        title="Vector", description="Vector of floats"), PropertyType.SYSTEM]
    set: Annotated[DescriptorSetDataModel, Field(
        title="Set", description="Descriptor set"), PropertyType.SYSTEM]


class PolygonDataModel(IdentityDataModel):
    """Polygon data model for ApertureDB.
    """
    type = ObjectType.POLYGON


class FrameDataModel(IdentityDataModel):
    """Frame data model for ApertureDB.
    """
    type = ObjectType.FRAME


class DescriptorSetDataModel(IdentityDataModel):
    """Descriptor Set data model for ApertureDB.
    """
    type = ObjectType.DESCRIPTORSET
    name: Annotated[str, Field(title="Name", description="Name of the descriptor set"),
                    PropertyType.SYSTEM]
    dimensions: Annotated[int, Field(title="Dimension", description="Dimension of the descriptor set"),
                          PropertyType.SYSTEM]


class BoundingBoxDataModel(IdentityDataModel):
    """Bounding Box data model for ApertureDB.
    """
    type = ObjectType.BOUNDING_BOX



import json
from types import SimpleNamespace
from aperturedb.ConnectorRest import ConnectorRest
from requests.sessions import Session


class TestUserConvenience():
    """
    This class tests some undocumented features of the Python SDK.
    This cannot rely on dbinfo, or connect as dbinfo and common lib rely on explicit
    arguments.
    """

    def test_ConnectorRest_handlesNonePort(self):
        """
        Test that ConnectorRest can handle a None port,
        and will default to the correct port.
        """
        client = ConnectorRest(host="dummy", user="admin", password="password")
        assert "443" in client.url
        posts = 0

        def mock_post(self, url, headers, files, verify):
            nonlocal posts
            assert "443" in url
            response1 = {
                "json": [{"Authenticate": {
                    "status": 0,
                    "session_token": "x",
                    "refresh_token": "2",
                    "session_token_expires_in": 3600,
                    "refresh_token_expires_in": 3600
                }}],
                "blobs": []
            }

            r = SimpleNamespace(status_code=200, text=json.dumps(response1))
            posts += 1
            return r
        old_post = Session.post
        Session.post = mock_post
        client.query("[{\"FindEntity\": {\"_ref\": 1}}]")
        # Ensure that the mock post was called, 1 time to authenticate, 1 time to query
        assert posts == 2
        Session.post = old_post



#!/usr/bin/env python3
import logging
import datetime
import os
import json
import requests
from string import Template
import platform
import faulthandler
import signal

# https://docs.python.org/3/library/faulthandler.html
# Register SIGUSR1 to dump the stack trace
# Good for debugging a running process
faulthandler.register(signal.SIGUSR1.value)

logger = logging.getLogger(__name__)

__version__ = "0.4.42"

# set log level
formatter = logging.Formatter(
    "%(asctime)s : %(levelname)s : %(name)s : %(thread)d : %(lineno)d : %(message)s")

log_file_level = logging.getLevelName(os.getenv("LOG_FILE_LEVEL", "WARN"))
log_console_level = logging.getLevelName(
    os.getenv("LOG_CONSOLE_LEVEL", "ERROR"))

# Set the logger filter to the minimum (more chatty) of the two handler levels
# This reduces problems if the environment adds a root handler (e.g. Google Colab)
logger_level = min(log_file_level, log_console_level)
if any(log_control in os.environ
       for log_control in ["LOG_CONSOLE_LEVEL", "LOG_FILE_LEVEL"]):
    logger.setLevel(logger_level)

# define file handler and set formatter
error_file_name = "error.${now}.log"

if "ADB_LOG_FILE" in os.environ:
    error_file_name = None if len(
        os.environ["ADB_LOG_FILE"]) == 0 else os.environ["ADB_LOG_FILE"]

if error_file_name is not None:
    error_file_tmpl = Template(error_file_name)
    template_items = {
        # python isodate has ':', not valid in files in windows.
        "now": str(datetime.datetime.now().isoformat()).replace(':', ''),
        "node": str(platform.node())
    }
    error_file_handler = logging.FileHandler(error_file_tmpl.safe_substitute(
        **template_items), delay=True)
    error_file_handler.setFormatter(formatter)
    error_file_handler.setLevel(log_file_level)
    logger.addHandler(error_file_handler)

error_console_handler = logging.StreamHandler()
error_console_handler.setLevel(log_console_level)
error_console_handler.setFormatter(formatter)
logger.addHandler(error_console_handler)

try:
    latest_version = json.loads(requests.get(
        "https://pypi.org/pypi/aperturedb/json").text)["info"]["version"]
except Exception as e:
    logger.warning(
        f"Failed to get latest version: {e}. You are using version {__version__}")
    latest_version = None
if __version__ != latest_version:
    logger.warning(
        f"The latest version of aperturedb is {latest_version}. You are using version {__version__}. It is recommended to upgrade.")



from pathlib import Path
import os
import subprocess
import runpy
import requests
import shutil
import pytest
import numpy as np
import pandas as pd
import os.path as osp
import tempfile
from aperturedb.Query import QueryBuilder, Query
from aperturedb.Entities import Entities
from aperturedb.Constraints import Constraints
from aperturedb.Images import Images
from aperturedb.Utils import Utils
from aperturedb.SPARQL import SPARQL
from aperturedb.cli.ingest import from_csv, TransformerType, IngestType
from aperturedb.ImageDataCSV import ImageDataCSV
from aperturedb.EntityDataCSV import EntityDataCSV
from aperturedb.ConnectionDataCSV import ConnectionDataCSV
from aperturedb.DescriptorDataCSV import DescriptorDataCSV
from aperturedb.ParallelLoader import ParallelLoader
from aperturedb.transformers.common_properties import CommonProperties
from aperturedb.transformers.image_properties import ImageProperties
from aperturedb.transformers.clip_pytorch_embeddings import CLIPPyTorchEmbeddings
from aperturedb.transformers.facenet_pytorch_embeddings import FacenetPyTorchEmbeddings

import logging
logger = logging.getLogger(__name__)


@pytest.fixture
def load_cookbook(utils: Utils, db):
    utils.remove_all_indexes()
    utils.remove_all_objects()

    temp_dir = tempfile.mkdtemp()
    # temp_path = Path(temp_dir)
    original_dir = os.getcwd()
    os.chdir(temp_dir)

    # Define the URL and file path for the script
    file_url = "https://raw.githubusercontent.com/aperture-data/Cookbook/refs/heads/main/scripts/convert_ingredients_adb_csv.py"
    file_path = Path("convert_ingredients_adb_csv.py")

    try:
        # Download the script file
        response = requests.get(file_url)
        file_path.write_text(response.text)

        runpy.run_path(str(file_path), run_name="__main__")

        data = ImageDataCSV("dishes.adb.csv")
        data = CLIPPyTorchEmbeddings(data, client=db)
        data = ImageProperties(data, client=db)
        data = CommonProperties(data, client=db)
        loader = ParallelLoader(db)
        loader.ingest(data, batchsize=100, stats=True)

        data = EntityDataCSV("ingredients.adb.csv")
        loader = ParallelLoader(db)
        loader.ingest(data, batchsize=100, stats=True)

        data = ConnectionDataCSV("dish_ingredients.adb.csv")
        loader = ParallelLoader(db)
        loader.ingest(data, batchsize=100, stats=True)
    finally:
        os.chdir(original_dir)


# Tag the test functions that depend on the setup as external_network
def pytest_collection_modifyitems(items):
    for item in items:
        if "load_cookbook" in getattr(item, "fixturenames", ()):
            item.add_marker("external_network")

# Test functions that depends on the setup


@pytest.fixture
def sparql(db):
    sparql = SPARQL(db)
    print(sparql.schema)
    assert sparql.connections, f"No connections {sparql.schema}"
    assert sparql.properties, f"No properties {sparql.schema}"
    return sparql


@pytest.mark.parametrize("description,query", [
    ('Find all images with chicken and butter as ingredients',
     """
SELECT ?s ?caption {
?s c:HasIngredient [p:name "chicken"] , [p:name "butter"] ;
    p:caption ?caption .
} LIMIT 10
"""),
    ('Find all images with chicken or butter as ingredients',
     """
SELECT ?s ?caption WHERE {
VALUES ?ingredient { "chicken" "butter" }
?s c:HasIngredient [p:name ?ingredient] ;
    p:caption ?caption .
} LIMIT 10
"""),
    ('Find the top 10 ingredients',
     """
SELECT (COUNT(*) AS ?count) ?ingredient WHERE {
?s c:HasIngredient [p:name ?ingredient] .
} GROUP BY ?ingredient ORDER BY DESC(?count) LIMIT 10
"""),
    ('Do a descriptor search for a random image',
     f"""
SELECT ?i ?distance ?d ?caption WHERE {{
?d knn:similarTo [
    knn:set 'ViT-B/16' ;
    knn:k_neighbors 20 ;
    knn:vector "{SPARQL.encode_descriptor(np.random.rand(512))}" ;
    knn:distance ?distance
] ;
    c:ANY ?i . # Use fake connection because we can't say c:_DescriptorConnection
    ?i p:caption ?caption .
}}
""")])
def test_sparql(load_cookbook, sparql, query, description):
    results = sparql.query(query)
    assert results, f"No results for {description}"



from __future__ import annotations
from typing import List, Tuple
from aperturedb import Parallelizer
import numpy as np
import logging
import inspect


from aperturedb.DaskManager import DaskManager
from aperturedb.Connector import Connector
from aperturedb.types import Commands, Blobs, CommandResponses
from aperturedb.CommonLibrary import execute_query

logger = logging.getLogger(__name__)


class ParallelQuery(Parallelizer.Parallelizer):
    """
    **Parallel and Batch Querier for ApertureDB**

    This class provides the abstraction for partitioning data into batches,
    so that they may be processed using different threads.

    Args:
        client (Connector): The database connector.
        dry_run (bool, optional): Whether to run in dry run mode. Defaults to False.
    """

    # 0 is success, 2 is object exists
    success_statuses = [0, 2]

    @classmethod
    def setSuccessStatus(cls, statuses: list[int]):
        cls.success_statuses = statuses

    @classmethod
    def getSuccessStatus(cls):
        return cls.success_statuses

    def __init__(self, client: Connector, dry_run: bool = False):
        super().__init__()
        test_string = f"Connection test successful with {client.config}"
        try:
            _, _ = client.query([{"GetSchema": {}}], [])
            logger.info(test_string)
        except Exception as e:
            logger.error(test_string.replace("successful", "failed"))
            raise

        self.client = client.clone()

        self.dry_run = dry_run

        self.type = "query"

        self.responses = []

        self.commands_per_query = 1
        self.blobs_per_query = 0
        self.daskManager = None
        self.batch_command = execute_query

    def generate_batch(self, data: List[Tuple[Commands, Blobs]]) -> Tuple[Commands, Blobs]:
        """
        Here we flatten the individual queries to run them as
        a single query in a batch
        We also update the _ref values and connections refs.

        Args:
            data (list[tuple[Query, Blobs]]): The data to be batched.  Each tuple contains a list of commands and a list of blobs.

        Returns:
            commands (Commands): The batched commands.
            blobs (Blobs): The batched blobs.
        """
        def update_refs(batched_commands):
            updates = {}
            for i, cmd in enumerate(batched_commands):
                if isinstance(cmd, list):
                    # Only parallel queries will work.
                    break
                values = cmd[list(cmd.keys())[0]]
                if "_ref" in values:
                    updates[values["_ref"]] = i + 1
                    values["_ref"] = i + 1
                    assert values["_ref"] < 100000
                if "image_ref" in values:
                    values["image_ref"] = updates[values["image_ref"]]
                if "video_ref" in values:
                    values["video_ref"] = updates[values["video_ref"]]
                if "is_connected_to" in values:
                    if "ref" in values["is_connected_to"]:
                        values["is_connected_to"]["ref"] = updates[values["is_connected_to"]["ref"]]
                    for op in ["any", "all"]:
                        if op in values["is_connected_to"]:
                            for idx in range(len(values["is_connected_to"][op])):
                                if "ref" in values["is_connected_to"][op][idx]:
                                    values["is_connected_to"][op][idx]["ref"] = updates[values["is_connected_to"][op][idx]["ref"]]
                if "connect" in values and "ref" in values["connect"]:
                    values["connect"]["ref"] = updates[values["connect"]["ref"]]
                if "src" in values:
                    values["src"] = updates[values["src"]]
                if "dst" in values:
                    values["dst"] = updates[values["dst"]]
                if "ref" in values:
                    values["ref"] = updates[values["ref"]]

            return batched_commands

        q = update_refs([cmd for query in data for cmd in query[0]])
        blobs = [blob for query in data for blob in query[1]]

        return q, blobs

    def call_response_handler(self, q: Commands, blobs: Blobs, r: CommandResponses, b: Blobs):
        try:
            self.generator.response_handler(q, blobs, r, b)
        except BaseException as e:
            logger.exception(e)

    def do_batch(self, client: Connector, batch_start: int,  data: List[Tuple[Commands, Blobs]]) -> None:
        """
        Executes batch of queries and blobs in the database.

        Args:
            client (Connector): The database connector.
            data (list[tuple[Commands, Blobs]]): The data to be batched.  Each tuple contains a list of commands and a list of blobs.

        It also provides a way for invoking a user defined function to handle the
        responses of each of the queries executed. This function can be used to process
        the responses from each of the corresponding queries in [Parallelizer](/python_sdk/parallel_exec/Parallelizer)
        It will be called once per query, and it needs to have 4 parameters:
        - requests
        - input_blobs
        - responses
        - output_blobs
        Example usage:
        ``` python
            class MyQueries(QueryGenerator):
                def process_responses(requests, input_blobs, responses, output_blobs):
                    self.requests.extend(requests)
                    self.responses.extend(responses)
            loader = ParallelLoader(self.client)
            generator = MyQueries()
            loader.ingest(generator)
        ```
        """

        q, blobs = self.generate_batch(data)

        query_time = 0
        worker_stats = {}
        if not self.dry_run:
            response_handler = None
            strict_response_validation = False
            if hasattr(self.generator, "response_handler") and callable(self.generator.response_handler):
                response_handler = self.generator.response_handler
            if hasattr(self.generator, "strict_response_validation") and isinstance(self.generator.strict_response_validation, bool):
                strict_response_validation = self.generator.strict_response_validation

            # if response_handler doesn't support index, just discard the index with a wrapper.
            if response_handler is not None:
                parameter_count = len(inspect.signature(
                    response_handler).parameters)
                if parameter_count < 4 or parameter_count > 5:
                    raise Exception("Bad Signature for response_handler :"
                                    f"expected 6 > args > 3, got {parameter_count}")
                if parameter_count == 4:
                    indexless_handler = response_handler
                    def response_handler(query, qblobs, resp, rblobs, qindex): return indexless_handler(
                        query, qblobs, resp, rblobs)

            result, r, b = self.batch_command(
                client,
                q,
                blobs,
                ParallelQuery.success_statuses,
                response_handler,
                self.commands_per_query,
                self.blobs_per_query,
                strict_response_validation=strict_response_validation,
                cmd_index=batch_start)
            if result == 0:
                query_time = client.get_last_query_time()
                worker_stats["succeeded_commands"] = len(q)
                worker_stats["succeeded_queries"] = len(data)
                worker_stats["objects_existed"] = sum(
                    [v['status'] == 2 for i in r for k, v in i.items()])
            elif result == 1:
                self.error_counter += 1
                worker_stats["succeeded_queries"] = 0
                worker_stats["succeeded_commands"] = 0
                worker_stats["objects_existed"] = 0
            elif result == 2:
                # with result 2, some queries might have failed.
                def filter_per_group(group):
                    return group.items() if isinstance(group, dict) else {}
                worker_stats["succeeded_commands"] = sum(
                    [v['status'] == 0 for i in r for k, v in filter_per_group(i)])
                worker_stats["objects_existed"] = sum(
                    [v['status'] == 2 for i in r for k, v in filter_per_group(i)])
                sq = 0
                for i in range(0, len(r), self.commands_per_query):
                    # Some errors stop the whole query from being executed
                    # https://docs.aperturedata.io/query_language/Overview/Responses#return-status
                    if issubclass(type(r), list):
                        if all([v['status'] == 0 for j in r[i:i + self.commands_per_query] for k, v in filter_per_group(j)]):
                            sq += 1
                worker_stats["succeeded_queries"] = sq
        else:
            query_time = 1

        # append is thread-safe
        self.times_arr.append(query_time)
        self.actual_stats.append(worker_stats)

    def worker(self, thid: int, generator, start: int, end: int):
        # A new connection will be created for each thread
        client = self.client.clone()

        total_batches = (end - start) // self.batchsize

        if (end - start) % self.batchsize > 0:
            total_batches += 1

        logger.info(f"Worker {thid} executing {total_batches} batches")
        for i in range(total_batches):

            batch_start = start + i * self.batchsize
            batch_end = min(batch_start + self.batchsize, end)

            try:
                self.do_batch(client, batch_start,
                              generator[batch_start:batch_end])
            except Exception as e:
                logger.exception(e)
                logger.warning(
                    f"Worker {thid} failed to execute batch {i}: [{batch_start},{batch_end}]")
                self.error_counter += 1

            if thid == 0 and self.stats:
                self.pb.update((i + 1) / total_batches)

    def get_objects_existed(self) -> int:
        return sum([stat["objects_existed"]
                    for stat in self.actual_stats])

    def get_succeeded_queries(self) -> int:
        return sum([stat["succeeded_queries"]
                    for stat in self.actual_stats])

    def get_succeeded_commands(self) -> int:
        return sum([stat["succeeded_commands"]
                    for stat in self.actual_stats])

    def query(self, generator, batchsize: int = 1, numthreads: int = 4, stats: bool = False) -> None:
        """
        This function takes as input the data to be executed in specified number of threads.
        The generator yields a tuple : (array of commands, array of blobs)
        Args:
            generator (_type_): The class that generates the queries to be executed.
            batchsize (int, optional): Number of queries per transaction. Defaults to 1.
            numthreads (int, optional): Number of parallel workers. Defaults to 4.
            stats (bool, optional): Show statistics at end of ingestion. Defaults to False.
        """

        use_dask = hasattr(generator, "use_dask") and generator.use_dask
        if use_dask:
            self._reset(batchsize=batchsize, numthreads=numthreads)
            self.daskmanager = DaskManager(num_workers=numthreads)

        if hasattr(self, "query_setup"):
            self.query_setup(generator)

        if use_dask:
            results, self.total_actions_time = self.daskmanager.run(
                self.__class__, self.client, generator, batchsize, stats=stats)
            self.actual_stats = []
            for result in results:
                if result is not None:
                    self.times_arr.extend(result.times_arr)
                    self.error_counter += result.error_counter
                    self.actual_stats.append(
                        {"succeeded_queries": result.succeeded_queries,
                         "succeeded_commands": result.succeeded_commands,
                         "objects_existed": result.objects_existed})
            self.total_actions = len(generator.df)

            if stats:
                self.print_stats()
        else:
            # allow subclass to do verification
            if issubclass(type(self), ParallelQuery) and hasattr(self, 'verify_generator') and callable(self.verify_generator):
                self.verify_generator(generator)
            elif len(generator) > 0:
                if isinstance(generator[0], tuple) and isinstance(generator[0][0], list):
                    # if len(generator[0]) > 0:
                    #
                    #  Not applicable to old style loaders.
                    self.commands_per_query = len(generator[0][0])
                    if len(generator[0][1]):
                        self.blobs_per_query = len(generator[0][1])
                else:
                    logger.error(
                        f"Could not determine query structure from:\n{generator[0]}")
                    logger.error(type(generator[0]))
            logger.info(
                f"Commands per query = {self.commands_per_query}, Blobs per query = {self.blobs_per_query}"
            )
            self.run(generator, batchsize, numthreads, stats)

    def print_stats(self) -> None:

        times = np.array(self.times_arr)
        total_queries_exec = len(times)

        print("============ ApertureDB Parallel Query Stats ============")
        print(f"Total time (s): {self.total_actions_time}")
        print(f"Total queries executed: {total_queries_exec}")

        if total_queries_exec == 0:
            print("All queries failed!")

        else:
            mean = np.mean(times)
            std = np.std(times)
            tp = 0 if mean == 0 else 1 / mean * self.numthreads

            print(f"Avg Query time (s): {mean}")
            print(f"Query time std: {std}")
            print(f"Avg Query Throughput (q/s): {tp}")

            i_tp = self.total_actions / self.total_actions_time
            print(
                f"Overall insertion throughput ({self.type}/s): {i_tp if self.error_counter == 0 else 'NaN'}")

            if self.error_counter > 0:
                err_perc = 100 * self.error_counter / total_queries_exec
                print(f"Total errors encountered: {self.error_counter}")
                print(f"Errors (%): {err_perc}")

        print("=========================================================")

    def debug_sample(self, **kwargs) -> None:
        """
        Sample the data to be ingested for debugging purposes.
        """
        if "sample_count" in kwargs:
            sample_count = kwargs["sample_count"]
        else:
            sample_count = len(self.generator)

        if sample_count > 0:
            print("Sample of data to be ingested:")
            for i in range(sample_count):
                print(self.generator[i])


def execute_batch(q: Commands, blobs: Blobs, db: Connector, *args, **kwargs) -> Tuple[int, CommandResponses, Blobs]:
    from aperturedb.CommonLibrary import execute_query, issue_deprecation_warning
    issue_deprecation_warning(
        "ParallelQuery.execute_batch", "CommonLibrary.execute_query")
    return execute_query(db, q, blobs, *args, **kwargs)
